#include <cudnn.h>
#include <iostream>
#include <numeric>
#include <memory>
#include <stdlib.h>
#include <assert.h>
#include <vector>
#include <cudnn_cnn_train_v8.h>
#include <cudnn_cnn_infer_v8.h>


#define CUDNN_CALL(f)                                                      \
  if (cudnnStatus_t unclash_cxx_e = (f); unclash_cxx_e != CUDNN_STATUS_SUCCESS) \
    std::cerr << __func__ << " line: " << __LINE__ << " cuDNN: " << cudnnGetErrorString(unclash_cxx_e) << std::endl;

template <typename T>
class Tensor {
    std::vector<int> dims_;
    int size_;

    struct deleteCudaPtr {
        void operator()(T *p) const {
            cudaFree(p);
        }
    };

    std::shared_ptr<T> ptr_;

public:

    Tensor() {}

    Tensor(std::vector<int> dims) : dims_(dims) {
        T* tmp_ptr;
        size_ = std::accumulate(dims_.begin(), dims_.end(), 1, std::multiplies<int>());
        cudaMalloc(&tmp_ptr, sizeof(T) * size_);

        ptr_.reset(tmp_ptr, deleteCudaPtr());
    }

    T* begin() const { return ptr_.get(); }
    T* end()   const { return ptr_.get() + size_; }
    int size() const { return size_; }
    std::vector<int> dims() const { return dims_; }
};

template <typename T>
Tensor<T> zeros(std::vector<int> dims) {
    Tensor<T> tensor(dims);
    int size = tensor.size();
    cudaMemset(tensor.begin(), 0, size * sizeof(T));
    return tensor;
}

template <typename T>
Tensor<T> rand(std::vector<int> dims, int seed) {
    Tensor<T> tensor(dims);
    int size = tensor.size();
    T *h = (T *)malloc(size * sizeof(T));
    srand(seed);
    for (int i = 0; i < size; i++){
        if (rand() % 2 == 0)
            h[i] = (T) (rand() / (RAND_MAX));
        else
            h[i] = (T) (-rand() / (RAND_MAX));
    }
    cudaMemcpy(tensor.begin(), h, size * sizeof(T), cudaMemcpyHostToDevice);
    cudaDeviceSynchronize();
    free(h);
    return tensor;
}

class CuDNNCommonOp {
  public:
    CuDNNCommonOp(){};
    CuDNNCommonOp(cudnnFusedOps_t op_id) : plan_created_(false) {
      // New 'fused op' descriptor creation
      CUDNN_CALL(cudnnCreateFusedOpsPlan(&op_, op_id));
      CUDNN_CALL(cudnnCreateFusedOpsConstParamPack(&op_const_params_, op_id));
      CUDNN_CALL(cudnnCreateFusedOpsVariantParamPack(&op_variant_params_, op_id));
    }

    // Launch op
    void Execute(cudnnHandle_t cudnn_handle) {
      if (!plan_created_) std::cerr << "CuDNNCommonOp exec requested without a valid 'plan', need: <set const params>, GetWorkspaceSizeBytes(), Execute()." << std::endl;
      CUDNN_CALL(cudnnFusedOpsExecute(cudnn_handle, op_, op_variant_params_));
    }

    template <typename T>
    void SetOpConstParamDesc(cudnnFusedOpsConstParamLabel_t param_label, T *param_ptr) {
      CUDNN_CALL(cudnnSetFusedOpsConstParamPackAttribute(op_const_params_, param_label, param_ptr));
      // Setting a 'const param pack' value invalidates the plan
      plan_created_ = false;
    }

    template <typename T>
    void SetOpConstParamAttr(cudnnFusedOpsConstParamLabel_t param_label, T param) {
      CUDNN_CALL(cudnnSetFusedOpsConstParamPackAttribute(op_const_params_, param_label, &param));
      plan_created_ = false;
    }

    template <typename T>
    void SetOpConstParamAttr(const std::vector<cudnnFusedOpsConstParamLabel_t> &param_labels, T param) {
    for (auto param_label : param_labels)
      SetOpConstParamAttr(param_label, param);
    }

    template <typename T>
    void SetOpVariantParamAttrPtr(cudnnFusedOpsVariantParamLabel_t param_label, T *param_ptr) {
      CUDNN_CALL(cudnnSetFusedOpsVariantParamPackAttribute(op_variant_params_,
                                                          param_label, param_ptr));
    }

    template <typename T>
    void SetOpVariantParamAttrPtr(const std::vector<cudnnFusedOpsVariantParamLabel_t> &param_labels, const T *param_ptr) {
    for (auto param_label : param_labels)
      SetOpVariantParamAttrPtr(param_label, param_ptr);
    }

    size_t GetWorkspaceSizeInBytes(cudnnHandle_t cudnn_handle) {
      size_t workspace_bytes = 0U;
      CUDNN_CALL(cudnnMakeFusedOpsPlan(cudnn_handle, op_, op_const_params_, &workspace_bytes));
      plan_created_ = true;
      return workspace_bytes;
    }
  private:
    bool plan_created_;

    cudnnFusedOpsPlan_t op_;
    cudnnFusedOpsConstParamPack_t op_const_params_;
    cudnnFusedOpsVariantParamPack_t op_variant_params_;
};

template <typename DType>
class CuDNNNormConvolutionOp{

    Tensor<float> workspace;
    size_t fwd_workspace_byte_;

    cudnnHandle_t d_handle;

    cudnnTensorDescriptor_t in_desc_;
    cudnnTensorDescriptor_t equiv_scale_bias_desc_;
    cudnnFilterDescriptor_t filter_desc_;
    cudnnTensorDescriptor_t out_desc_;
    cudnnTensorDescriptor_t in_stats_desc_;
    cudnnTensorDescriptor_t out_stats_desc_;
    cudnnConvolutionDescriptor_t conv_desc_;

    bool fwd_op_plan_output_stats_;
    bool no_norm;

    cudnnFusedOpsPointerPlaceHolder_t fprop_eq_scale_bias_ptr_type_;

    DType *ones_ptr;
    DType *zeros_ptr;

    cudnnActivationDescriptor_t activation_desc_;
    cudnnDataType_t dtype_param_;

    CuDNNCommonOp fwd_op_;

  public:
    CuDNNNormConvolutionOp():
      fwd_op_(CUDNN_FUSED_SCALE_BIAS_ACTIVATION_CONV_BNSTATS)
    {
        CUDNN_CALL(cudnnCreate(&d_handle));
        CUDNN_CALL(cudnnCreateTensorDescriptor(&in_desc_));
        CUDNN_CALL(cudnnCreateTensorDescriptor(&equiv_scale_bias_desc_));
        CUDNN_CALL(cudnnCreateTensorDescriptor(&out_desc_));
        CUDNN_CALL(cudnnCreateTensorDescriptor(&in_stats_desc_));
        CUDNN_CALL(cudnnCreateTensorDescriptor(&out_stats_desc_));
        CUDNN_CALL(cudnnCreateFilterDescriptor(&filter_desc_));
        CUDNN_CALL(cudnnCreateConvolutionDescriptor(&conv_desc_));
        CUDNN_CALL(cudnnCreateActivationDescriptor(&activation_desc_));
    }

    void Init() {
        cudnnDataType_t dtype_ = CUDNN_DATA_HALF;
        cudnnDataType_t cudnn_fwd_compute_type = CUDNN_DATA_FLOAT;
        cudnnTensorFormat_t format_ = CUDNN_TENSOR_NHWC;
        std::vector<int> dshape {128,64,56,56};
        std::vector<int> dstride {200704,1,3584,64};
        
        std::vector<int> oshape {128,256,56,56};
        std::vector<int> ostride {802816,1,14336,256};

        std::vector<int> equiv_scale_shape = {1, 64, 1, 1};
        std::vector<int> equiv_scale_stride = {64, 1, 1, 1};

        std::vector<int> stats_shape = {1, 256, 1, 1};
        std::vector<int> stats_stride = {256, 1, 1, 1};

        size_t equiv_scale_bytes = 64 * sizeof(DType);
        cudaMemsetAsync(&zeros_ptr, 0,equiv_scale_bytes);
        cudaMemsetAsync(&ones_ptr, 1,equiv_scale_bytes);

        size_t workspace_bytes = 0U;

        fwd_op_.SetOpConstParamAttr({CUDNN_PARAM_XDATA_PLACEHOLDER,
                                    CUDNN_PARAM_WDATA_PLACEHOLDER,
                                    CUDNN_PARAM_YDATA_PLACEHOLDER}, CUDNN_PTR_16B_ALIGNED);
        fwd_op_.SetOpConstParamAttr({CUDNN_PARAM_BN_EQSCALE_PLACEHOLDER,
                                    CUDNN_PARAM_BN_EQBIAS_PLACEHOLDER}, CUDNN_PTR_16B_ALIGNED);
        fwd_op_.SetOpConstParamAttr({CUDNN_PARAM_YSUM_PLACEHOLDER,
                                    CUDNN_PARAM_YSQSUM_PLACEHOLDER}, CUDNN_PTR_16B_ALIGNED);
        
        cudnnSetConvolution2dDescriptor(conv_desc_, 0, 0, 1, 1, 1, 1, CUDNN_CROSS_CORRELATION, cudnn_fwd_compute_type);

        cudnnSetFilter4dDescriptor(filter_desc_, dtype_, format_, 256, 64, 1, 1);

        cudnnSetConvolutionMathType(conv_desc_, CUDNN_TENSOR_OP_MATH);

        cudnnSetConvolutionGroupCount(conv_desc_, 1);

        cudnnSetTensorNdDescriptor(in_desc_, dtype_, 4, dshape.data(), dstride.data());

        cudnnSetTensorNdDescriptor(out_desc_, dtype_, 4, oshape.data(), ostride.data());

        cudnnSetTensorNdDescriptor(equiv_scale_bias_desc_, dtype_, 4, equiv_scale_shape.data(), equiv_scale_stride.data());

        fwd_op_.SetOpConstParamDesc(CUDNN_PARAM_BN_EQSCALEBIAS_DESC, equiv_scale_bias_desc_);

        cudnnSetTensorNdDescriptor(out_stats_desc_, cudnn_fwd_compute_type, 4, stats_shape.data(), stats_stride.data());

        fwd_op_.SetOpConstParamDesc(CUDNN_PARAM_YSTATS_DESC, out_stats_desc_);

        fwd_op_.SetOpConstParamDesc(CUDNN_PARAM_XDESC, in_desc_);

        fwd_op_.SetOpConstParamAttr(CUDNN_PARAM_BN_MODE, CUDNN_BATCHNORM_SPATIAL);

        fwd_op_.SetOpConstParamDesc(CUDNN_PARAM_CONV_DESC, conv_desc_);

        fwd_op_.SetOpConstParamDesc(CUDNN_PARAM_WDESC, filter_desc_);

        fwd_op_.SetOpConstParamDesc(CUDNN_PARAM_YDESC, out_desc_);

        fwd_workspace_byte_ = fwd_op_.GetWorkspaceSizeInBytes(d_handle);
        std::vector<int> shape {static_cast<int>(fwd_workspace_byte_/sizeof(float)), 1};
        workspace = zeros<float>(shape);
    }

    void Forward(const Tensor<DType> &in_data,
                const Tensor<DType> &weight,
                const Tensor<DType> &out_data,
                const Tensor<float> &out_sum,
                const Tensor<float> &out_sum_of_squares) {
        fwd_op_.SetOpVariantParamAttrPtr(CUDNN_PTR_XDATA, in_data.begin());
        fwd_op_.SetOpVariantParamAttrPtr(CUDNN_PTR_WDATA, weight.begin());
        fwd_op_.SetOpVariantParamAttrPtr(CUDNN_PTR_BN_EQSCALE, &ones_ptr);
        fwd_op_.SetOpVariantParamAttrPtr(CUDNN_PTR_BN_EQBIAS, &zeros_ptr);
        fwd_op_.SetOpVariantParamAttrPtr(CUDNN_PTR_WORKSPACE, workspace.begin());
        fwd_op_.SetOpVariantParamAttrPtr(CUDNN_SCALAR_SIZE_T_WORKSPACE_SIZE_IN_BYTES, &fwd_workspace_byte_);
        fwd_op_.SetOpVariantParamAttrPtr(CUDNN_PTR_YDATA, out_data.begin());
        fwd_op_.SetOpVariantParamAttrPtr(CUDNN_PTR_YSUM, out_sum.begin());
        fwd_op_.SetOpVariantParamAttrPtr(CUDNN_PTR_YSQSUM, out_sum_of_squares.begin());
        
        fwd_op_.Execute(d_handle);
    }
};